/*
 *
 *
 *										    
 * Sm.c (State Machine)
 * 
 * Written for EagleIV driver: Copyright 2006 Ikanos Communication
 *
 * This file is part of the eagle-usb driver package.
 *
 * eagle-usb driver package is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * "eagle-usb driver package" is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with "ADI USB ADSL Driver for Linux"; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: Sm.c,v 1.6 2004/09/27 21:59:07 mcoolive Exp $
 */

#include "Adiutil.h"
#include "eu_sm.h"
#include "eu_eth.h"
#include "Cmv.h"
#include "macros.h"
#include "eu_utils.h"
#include "eu_msg.h"
#include "Me.h"
#include "debug.h"
#include "eu_boot_sm.h"
#ifdef LINUX_2_4
#include <linux/tqueue.h>
#elif defined(LINUX_2_6)
#include <linux/workqueue.h>
#else
#warning "Unsupported version"
#endif

#define RESET_TIMER(_TIMEOUT_) pAdiSM->ModemReplyExpected = MP_FUNCTION_TYPE_WAITING_4_REPLY; \
                               SetTimerInterval(ins,_TIMEOUT_); 

#define ENDCASE SetTimerInterval(ins,TRANS_TIME_50MSEC_PAUSE); break;

#define OPER_CHECK_FREQ           3
#define OPER_CHECK_FREQ_DETAIL  0xF

#define LOS_DEFECT_RETRAIN_THRESHOLD    2   /* about  1  */
#define CRC_RETRAIN_THRESHOLD           4   /* about 10 secs of continuous CRC to cause retrain */


#define RETRY_LIMIT             10
#define MAX_IDLE_COUNT          100
#define RETRAIN_ATTEMPTS        160


/* quasi-functions to force beginning of SM reset   */
/* actual processing would happen when timer expires*/
#define TimerResetModemSM SetTimerInterval(ins,TRANS_TIME_2_BEGINRESET); \
                          ResetModemSM(pAdiSM);

#ifdef LINUX_2_4
static struct tq_struct eth_create_cb = {
	routine: eu_eth_create,
};
#endif


/*
FFD_DEBUG: EIV SM ProtoTypes
*/

void e4_StateReset(eu_instance_t *ins);
void e4_StateSetup(eu_instance_t *ins);
void e4_StateUntrained(eu_instance_t *ins);
void e4_StateOperational(eu_instance_t *ins);




void e4_ModemSM(eu_instance_t *ins)
{
	
	EIV_STATE *pCurrState = &ins->e4sm.currState;
	
	/*
	EIV_STATE *pOldState = &ins->e4sm.oldState;
	EIV_SUBSTATE *pSubState = &ins->e4sm.subState;
	*/
	/*eu_enters (DBG_SM);*/
	
     /*
     * Create ethernet device if not present
     */

    if ( ( ins->eth == NULL ) &&
         ( ! EU_TEST_FLAG (ins, EU_ETH_REGISTERING ) ) &&
         ( (*pCurrState) == EIV_STATE_OPERATIONAL ) )
    {
        EU_SET_FLAG (ins,EU_ETH_REGISTERING);

#ifdef LINUX_2_4
        INIT_TQUEUE (&eth_create_cb, eu_eth_create, ins);        
        if ( !schedule_task (  &eth_create_cb ) ) 
#elif defined(LINUX_2_6)

        if ( !schedule_work( &ins->create_eth) )
#else
        #warning "Unsupported version"
#endif
        {
            eu_err ("Sm: Scheduled net dev. creation dropped.\n");
            EU_CLEAR_FLAG (ins,EU_ETH_REGISTERING);
        }
        else
        {
            eu_dbg (DBG_SM,"Ethernet dev. creation scheduled.\n");
	    SetTimerInterval(ins,1000);
	    return; // we return here to give some time to the task finish before reading any other cmv 
        }
    }


    if (ins->eth) 
    {
        if ( ( (*pCurrState) == EIV_STATE_OPERATIONAL ) &&
             ( !netif_carrier_ok(ins->eth) ) 
           ) 
        {
            eu_dbg (DBG_SM,"** Carrier ON ** \n");
	    netif_carrier_on(ins->eth);
        }
        else if ( (*pCurrState) != EIV_STATE_OPERATIONAL )
        {
            eu_dbg (DBG_SM,"** Carrier OFF **\n");
            netif_carrier_off(ins->eth);
        }
    }
    
    
    switch ( *pCurrState )
    {
	case EIV_STATE_RESET:
		e4_StateReset(ins); 
		break;
	case EIV_STATE_SETUP:
		e4_StateSetup(ins);
		break;
	case EIV_STATE_UNTRAINED:
		e4_StateUntrained(ins);
		break;
	case EIV_STATE_OPERATIONAL:
		e4_StateOperational(ins);
		break;
	default:
		break;
     }
	/*eu_leaves (DBG_SM);*/
}

void e4_StateReset(eu_instance_t *ins)
{
	
	/*
	unsigned short int usModemReadyMsg[EAGLEIV_MP_BUFFER_SIZE]; 
	unsigned int decode;
	EIV_STATE *pOldState = &ins->e4sm.oldState;
	*/
	
	static unsigned short int cntrModemReady;
	EIV_STATE *pCurrState = &ins->e4sm.currState;
	EIV_SUBSTATE *pSubState = &ins->e4sm.subState;
	EIV_STAT *pStat = &ins->e4sm.stat;
	
	eu_enters (DBG_SM);	
	eu_dbg (DBG_SM,"e4_StateReset: %d\n",*pSubState);
	switch (*pSubState)
	{
	case EIV_SUBSTATE_INIT:
		
		ins->e4sm.mp_rx_async_msg_pending = false; 
		ins->e4sm.mp_tx_msg_pending = false;
		cntrModemReady = 0;
		
		
		/*
		if ( !EU_TEST_FLAG(ins,EU_HAS_INT) || boot_the_e4modem (ins) != 0)
                    eu_dbg (DBG_SM,"e4_StateReset: in substate INIT, Unable to load DSP code to device!\n");
                */
		if ( EU_TEST_FLAG(ins,EU_HAS_INT) )
                {
			ins->swap_data = 0xFF;
			EU_SCHEDULE ( &ins->e4_loadp);
			*pSubState = EIV_SUBSTATE_RX;
		}
                /*else
		    *pSubState = EIV_SUBSTATE_RX;
		*/
		    	    
	break;
	case EIV_SUBSTATE_TX:
		
		/* 
		*we should not come to this state because we are
		*waiting for ASYNC message : MODEM READY FROM L1
		*/
		
		break;
	case EIV_SUBSTATE_RX:
		if (ins->e4sm.mp_rx_async_msg_pending == true) //We got the ASYNC Msg 
		{
		      /*
		      *FFD_Debug: should decode the message and confirm it is modem Ready but since
		      *we only get one Asynch CMV from L1, that should be modem ready!
		      *right now decode returns success all the time but we need to handle this 
		      *when we change the decode function.
		      */
		      
		      eu_dbg (DBG_SM,"SM.Reset.init: GOT ASYN CMV => MODEM READY\n");
		     *pSubState = EIV_SUBSTATE_DONE;
		        
		}
		else
		{
			cntrModemReady++;
			if ( cntrModemReady > 100 )
			{
				*pSubState = EIV_SUBSTATE_INIT;
				cntrModemReady = 0;
				pStat->cntrModemReadyFailure++;
			}

		}

		break;
	case EIV_SUBSTATE_DONE:
		*pCurrState = EIV_STATE_SETUP;
		*pSubState = EIV_SUBSTATE_INIT;
		SetTimerInterval(ins,2000);
		break;
	default:
		 eu_dbg (DBG_SM,"** STATE MACHINE ERROR -> RESET STATE **\n");
		break;
	}

	eu_leaves (DBG_SM);

}



void e4_StateSetup(eu_instance_t *ins)
{
	unsigned short int *usCMVResponse; 
	unsigned int decode;
	static unsigned int uiCntr, cntrTXPending, cntrRXPending;
	unsigned int uiMaxNumOfCMV;
	EIV_STATE *pCurrState = &ins->e4sm.currState;
	/*EIV_STATE *pOldState = &ins->e4sm.oldState;*/
	EIV_SUBSTATE *pSubState = &ins->e4sm.subState;
	EIV_STAT *pStat = &ins->e4sm.stat;
	uiMaxNumOfCMV =ins->e4sm.usNumSetupCmvs;
	
	eu_enters (DBG_SM);	

	switch (*pSubState)
	{
	case EIV_SUBSTATE_INIT:
		eu_dbg (DBG_SM,"uiMaxNumOfCMV = %d\n",uiMaxNumOfCMV);
		cntrTXPending = 0;
		cntrRXPending = 0;
		uiCntr=0;
		ins->e4sm.mp_rx_sync_msg_pending = false; 
		ins->e4sm.mp_tx_msg_pending = false;
		*pSubState = EIV_SUBSTATE_TX;
 
	break;
	case EIV_SUBSTATE_TX:
	
		
		eu_dbg (DBG_SM,"StateSetup_TX>>MSG = %04X %04X %04X %04X %04X %04X %04X %04X>\n",
			ins->e4sm.mp_retrain_sequence[uiCntr][0], ins->e4sm.mp_retrain_sequence[uiCntr][1], 
			ins->e4sm.mp_retrain_sequence[uiCntr][2], ins->e4sm.mp_retrain_sequence[uiCntr][3],
			ins->e4sm.mp_retrain_sequence[uiCntr][4], ins->e4sm.mp_retrain_sequence[uiCntr][5], 
			ins->e4sm.mp_retrain_sequence[uiCntr][6], ins->e4sm.mp_retrain_sequence[uiCntr][7]);
			
		if ( eu_send_msg(ins, &ins->e4sm.mp_retrain_sequence[uiCntr][0]) == true )

			*pSubState = EIV_SUBSTATE_RX; 
		else
		{
			cntrTXPending++;
			if ( cntrTXPending > 20 )
			{
				pStat->cntrSetupTXPending++;
				cntrTXPending = 0;
				*pSubState = EIV_SUBSTATE_INIT;
				*pCurrState = EIV_STATE_RESET;
				eu_dbg (DBG_SM," SETUP TX Pending -> reseting the modem.\n");

			}
			eu_dbg (DBG_SM,"Could not write the CMV in EagleIV_StateSetup in EIV_SUBSTATE_TX\n");
		}
		
		break;

	case EIV_SUBSTATE_RX:
		cntrTXPending = 0;
		if (ins->e4sm.mp_rx_sync_msg_pending == true) 
		{
			/* decode the message into mp_rx_msg */
			if ( (decode = e4_mp_msg_decode(&ins->e4sm.e4_rx_cmv, ins->e4sm.raw_rx_msg_sync)) == MP_DECODE_OK) 
				*pSubState = EIV_SUBSTATE_DONE;
			else
			{
				*pSubState = EIV_SUBSTATE_INIT;
				pStat->cntrCMVFailure++;
			}
			usCMVResponse = (unsigned short int *)&ins->e4sm.e4_rx_cmv;
			eu_dbg (DBG_SM,"StateSetup_RX>>SYNC MSG = %04X %04X %04X %04X %04X %04X %04X %04X <decode = %08X>\n",
			usCMVResponse[0], usCMVResponse[1], usCMVResponse[2], usCMVResponse[3],
			usCMVResponse[4], usCMVResponse[5], usCMVResponse[6], usCMVResponse[7], decode);
			ins->e4sm.mp_rx_sync_msg_pending = false;
			
		}
		else
		{
			cntrRXPending++;
			if ( cntrRXPending > 20 )
			{
				pStat->cntrSetupRXPending++;
				cntrRXPending = 0;
				*pSubState = EIV_SUBSTATE_INIT;
				*pCurrState = EIV_STATE_RESET;
				eu_dbg (DBG_SM,"SETUP RX Pending -> reseting the modem.\n");

			}
			eu_dbg (DBG_SM,"SETUP : EIV_SUBSTATE_RX : get fail\n");

		}
		
		ENDCASE
		
	case EIV_SUBSTATE_DONE:
		cntrRXPending = 0;
		uiCntr++;
		if (uiCntr<uiMaxNumOfCMV)
		{
			*pSubState = EIV_SUBSTATE_TX;
		}
		else
		{
			
			*pCurrState = EIV_STATE_UNTRAINED;
			*pSubState = EIV_SUBSTATE_INIT;
			eu_dbg (DBG_SM,"SETUP : EIV_SUBSTATE_DONE: cntr <%d> going to Untrained State\n",uiCntr);
			SetTimerInterval(ins,30000);
			break;
		}

		ENDCASE

	default:
		eu_dbg (DBG_SM,"in e4_StateSetup => invalid substate\n");
		break;
	}
	
	eu_leaves (DBG_SM);

}

	
void e4_StateUntrained(eu_instance_t *ins)
{
	unsigned short int *usCMVResponse; 
	unsigned int decode;
	static unsigned int uiCntr, cntrTXPending, cntrRXPending;
	EIV_STATE *pCurrState = &ins->e4sm.currState;
	/*EIV_STATE *pOldState = &ins->e4sm.oldState;*/
	EIV_SUBSTATE *pSubState = &ins->e4sm.subState;
	EIV_STAT *pStat = &ins->e4sm.stat;

	eu_enters (DBG_SM);

	switch (*pSubState)
	{
	case EIV_SUBSTATE_INIT:
		cntrTXPending = 0;
		cntrRXPending = 0;
		uiCntr=0;
		ins->e4sm.mp_rx_sync_msg_pending = false; 
		ins->e4sm.mp_tx_msg_pending = false;
		
		e4_mp_msg_encode('r',cmd_Group2, 0  ,0 , 0x0,1,&ins->e4sm.mp_statcmv[0]);
		
		*pSubState = EIV_SUBSTATE_TX;
 
	break;
	case EIV_SUBSTATE_TX:
			
		if ( eu_send_msg(ins, &ins->e4sm.mp_statcmv[0]) == true )

			*pSubState = EIV_SUBSTATE_RX; 
		else
		{
			cntrTXPending++;
			if ( cntrTXPending > 20 )
			{
				pStat->cntrSetupTXPending++;
				cntrTXPending = 0;
				*pSubState = EIV_SUBSTATE_INIT;
				*pCurrState = EIV_STATE_RESET;
				eu_dbg (DBG_SM," Untrained TX Pending -> reseting the modem.\n");

			}
			eu_dbg (DBG_SM,"Could not read the CMV in e4_StateUntrained in EIV_SUBSTATE_TX\n");
		}
		
		break;
		
	case EIV_SUBSTATE_RX:
		cntrTXPending = 0;
		if (ins->e4sm.mp_rx_sync_msg_pending == true) 
		{
			/* decode the message into mp_rx_msg */
			if ( (decode = e4_mp_msg_decode(&ins->e4sm.e4_rx_cmv, ins->e4sm.raw_rx_msg_sync)) == MP_DECODE_OK)
			{ 
				*pSubState = EIV_SUBSTATE_DONE;
				ins->e4sm.usLinkStat = (uint16_t) (ins->e4sm.e4_rx_cmv.payload.ulong[0] & 0xFFFF);
			}
			else
			{
				*pSubState = EIV_SUBSTATE_INIT;
				pStat->cntrCMVFailure++;
			}
			usCMVResponse = (unsigned short int *)&ins->e4sm.e4_rx_cmv;
			eu_dbg (DBG_SM,"StateUntrained_RX>>SYNC MSG = %04X %04X %04X %04X %04X %04X %04X %04X <decode = %08X>\n",
			usCMVResponse[0], usCMVResponse[1], usCMVResponse[2], usCMVResponse[3],
			usCMVResponse[4], usCMVResponse[5], usCMVResponse[6], usCMVResponse[7], decode);
			ins->e4sm.mp_rx_sync_msg_pending = false;
			
		}
		else
		{
			cntrRXPending++;
			if ( cntrRXPending > 20 )
			{
				pStat->cntrSetupRXPending++;
				cntrRXPending = 0;
				*pSubState = EIV_SUBSTATE_INIT;
				*pCurrState = EIV_STATE_RESET;
				eu_dbg (DBG_SM,"Untrained RX Pending -> reseting the modem.\n");
				SetTimerInterval(ins,TRANS_TIME_4_RESETINIT);
				

			}
			eu_dbg (DBG_SM,"Untrained : EIV_SUBSTATE_RX : get fail\n");

		}
		
		break;
	case EIV_SUBSTATE_DONE:
		cntrRXPending = 0;
		eu_dbg (DBG_SM,"STAT00 %04X \n",ins->e4sm.usLinkStat);
		/* ins->e4sm.GUICMVList[EIV_CMDTAG_LINKSTAT] = (ULONG) ins->e4sm.usLinkStat;*/
		switch (ins->e4sm.usLinkStat)
		{
			case 0x0: case 0x1: case 0x3:	case 0x4:
				/* off */
				*pSubState = EIV_SUBSTATE_TX;
			break;

			case 0x2: /* FAIL*/
				*pSubState = EIV_SUBSTATE_INIT;
				*pCurrState = EIV_STATE_RESET;
				pStat->cntrLinkStatFaiure++;
			break;

			case 0x5: case 0x6: case 0x9:	case 0xa:
				*pSubState = EIV_SUBSTATE_TX;
			break;

			case 0x7:
				pStat->cntrShowtime++;
				
				*pCurrState = EIV_STATE_OPERATIONAL;
				*pSubState = EIV_SUBSTATE_INIT;
				SetTimerInterval(ins,1000); 
				 
			break;
			default:
				*pSubState = EIV_SUBSTATE_INIT;
			break;
		}

		break;

	default:
		eu_dbg (DBG_SM,"in e4_StateSetup => should not be here!!!\n");
		break;
	}

	eu_leaves (DBG_SM);
}

void e4_StateOperational(eu_instance_t *ins)
{

	unsigned short int *usCMVResponse; 
	unsigned int decode;
	static unsigned int uiCntr, cntrTXPending, cntrRXPending;
	unsigned int uiMaxNumOfCMV;
	uint32_t ulData[EAGLEIV_CMV_PAYLOAD_SIZE/2];
	EIV_STATE *pCurrState = &ins->e4sm.currState;
	/*EIV_STATE *pOldState = &ins->e4sm.oldState;*/
	EIV_SUBSTATE *pSubState = &ins->e4sm.subState;
	EIV_STAT *pStat = &ins->e4sm.stat;
	uiMaxNumOfCMV = ins->e4sm.usNumOperationalCmvs;
	
	/* eu_enters (DBG_SM);*/
        eu_dbg (DBG_SM,"e4_StateOperational: %d\n",*pSubState);

	switch (*pSubState)
	{
	case EIV_SUBSTATE_INIT:

		eu_cmd_to_modem(ins, EU_CMD_SET_MODE, MODE_LOOPBACK_OFF, 0, NULL); 
		wake_up_interruptible(&ins->sync_q); 
		uiCntr=0;
		*pSubState = EIV_SUBSTATE_TX;
 
	break;
	case EIV_SUBSTATE_TX:
		if ( eu_send_msg(ins, &ins->e4sm.opCMVS[uiCntr].usMP[0]) == true )
			*pSubState = EIV_SUBSTATE_RX; 
		else
		{
			cntrTXPending++;
			if ( cntrTXPending > 20 )
			{
				pStat->cntrOperationalTXPending++;
				cntrTXPending = 0;
				*pSubState = EIV_SUBSTATE_INIT;
				*pCurrState = EIV_STATE_RESET;
				eu_dbg (DBG_SM,"OPERATIONAL TX Pending -> reseting the modem.\n");

			}

			/* pStat->cntrOperationalTXPending */
			eu_dbg (DBG_SM,"Could not read OPERATIONAL CMV\n");
		}	
		break;

	case EIV_SUBSTATE_RX:
		cntrTXPending = 0;
		if (ins->e4sm.mp_rx_sync_msg_pending == true) 
		{
			/* decode the message into mp_rx_msg  */
			if ( (decode = e4_mp_msg_decode(&ins->e4sm.e4_rx_cmv, ins->e4sm.raw_rx_msg_sync)) == MP_DECODE_OK)
				*pSubState = EIV_SUBSTATE_DONE;
			else
			{
				pStat->cntrCMVFailure++;
				uiCntr++;
				*pSubState = EIV_SUBSTATE_TX;
				if (uiCntr>=uiMaxNumOfCMV)
				{
					eu_dbg (DBG_SM,"pStat->cntrCMVFailure<%d> reseting the cntr to 0\n", pStat->cntrCMVFailure);
					uiCntr = 0;
				}
			}
			usCMVResponse = (unsigned short int *)&ins->e4sm.e4_rx_cmv;
			eu_dbg (DBG_SM,"StateOperational_RX>>SYNC MSG = %04X %04X %04X %04X %04X %04X %04X %04X <decode = %08X>\n",
			usCMVResponse[0], usCMVResponse[1], usCMVResponse[2], usCMVResponse[3],
			usCMVResponse[4], usCMVResponse[5], usCMVResponse[6], usCMVResponse[7], decode);
			
		}
		else
		{
			cntrRXPending++;
			if ( cntrRXPending > 20 )
			{
				pStat->cntrOperationalRXPending++;
				cntrRXPending = 0;
				*pSubState = EIV_SUBSTATE_INIT;
				*pCurrState = EIV_STATE_RESET;
				eu_dbg (DBG_SM,"OPERATIONAL RX Pending -> reseting the modem.\n");
			}

			eu_dbg (DBG_SM,"cmv#%d => cmv error\n",uiCntr);
		}
		
		break;
	case EIV_SUBSTATE_DONE:
		cntrRXPending = 0;
		ulData[0] = ins->e4sm.e4_rx_cmv.payload.ulong[0];
		ulData[1] = ins->e4sm.e4_rx_cmv.payload.ulong[1];
		eu_dbg (DBG_SM,"cmd#[%d]=%.8X  %.8X\n",uiCntr,ulData[0],ulData[1]);

		switch (ins->e4sm.opCMVS[uiCntr].cmvTag)
		{
		
		case EIV_CMDTAG_UR:
			ins->e4sm.opCMVS[uiCntr].ulData[0] = ulData[0];
			ins->e4sm.opCMVS[uiCntr].ulData[1] = ulData[1];
			if ( ulData[0] != 0 )
			{
				
				ins->e4sm.GUICMVList[EIV_CMDTAG_CT] = API_CONN_TYPE_INTERLEAVED;
				ins->e4sm.GUICMVList[EIV_CMDTAG_UR] = ulData[0];
				ins->e4sm.UpstreamRate = ulData[0];
			}
			else
			{
				ins->e4sm.GUICMVList[EIV_CMDTAG_CT] = API_CONN_TYPE_FAST;
				ins->e4sm.GUICMVList[EIV_CMDTAG_UR] = ulData[1];
				ins->e4sm.UpstreamRate = ulData[0];
			}
			break;
		
		case EIV_CMDTAG_DR:
			ins->e4sm.ucRateUpdateFlag = 1;
			ins->e4sm.opCMVS[uiCntr].ulData[0] = ulData[0];
			ins->e4sm.opCMVS[uiCntr].ulData[1] = ulData[1];
			if ( ulData[0] != 0 )
			{
		
				ins->e4sm.GUICMVList[EIV_CMDTAG_DR] = ulData[0];
				ins->e4sm.DownstreamRate = ulData[0];
			}
			else
			{
		
				ins->e4sm.GUICMVList[EIV_CMDTAG_DR] = ulData[1];
				ins->e4sm.DownstreamRate = ulData[0];
			}
			
			if ( !EU_TEST_FLAG(ins,EU_LOW_RATE) )
                        {
                            if (ins->e4sm.DownstreamRate < (1024*1024))
                            {
                                eu_dbg (DBG_SM," **** LOW LINE RATE (0x%x) ****\n",ins->e4sm.DownstreamRate);
                                eu_cmd_to_modem(ins, EU_CMD_SET_TIMEOUT, 0, 0, NULL);
                            }
                            else						 
                            {
                                eu_dbg (DBG_SM,"***** HIGH LINE RATE (0x%x) *****\n", ins->e4sm.DownstreamRate);
                                
                                eu_cmd_to_modem(ins, EU_CMD_SET_TIMEOUT, 0xFF, 0, NULL);
                            }
                        }
                        
                        EU_SET_FLAG (ins, EU_LOW_RATE);
		
			break;
                
		case EIV_CMDTAG_FV:
			
			ins->e4sm.opCMVS[uiCntr].ulData[0] = ulData[0];
			ins->e4sm.opCMVS[uiCntr].ulData[1] = ulData[1];
			ins->e4sm.GUICMVList[EIV_CMDTAG_FV] = ulData[0];
			ins->e4sm.GUICMVList[EIV_CMDTAG_FD] = (ulData[0] & 0x0000FFFF) | (( ulData[1] << 16 ) & 0xFFFF0000);
			eu_dbg (DBG_SM,"FD: %.8X\n",ins->e4sm.GUICMVList[EIV_CMDTAG_FD]);

			break;
		
		case EIV_CMDTAG_AS:
			ins->e4sm.GUICMVList[EIV_CMDTAG_AS] = ulData[0] & 0x0000FFFF;
			ins->e4sm.opCMVS[uiCntr].ulData[0] = ulData[0];
			break;
		
		case EIV_CMDTAG_CV1:
			ins->e4sm.opCMVS[uiCntr].ulData[0] = ulData[0];
			ins->e4sm.opCMVS[uiCntr].ulData[1] = ulData[1];
			ins->e4sm.GUICMVList[EIV_CMDTAG_CV1] = ntohl(ulData[0]);
			ins->e4sm.GUICMVList[EIV_CMDTAG_CV2] = ntohl(ulData[1]);
			break;
			
		
		case EIV_CMDTAG_CPV1:
			ins->e4sm.opCMVS[uiCntr].ulData[0] = ulData[0];
			ins->e4sm.opCMVS[uiCntr].ulData[1] = ulData[1];
			ins->e4sm.GUICMVList[EIV_CMDTAG_CPV1] = ntohl(ulData[0]);
			ins->e4sm.GUICMVList[EIV_CMDTAG_CPV2] = ntohl(ulData[1]);
			break;
		
		case EIV_CMDTAG_NA:
			ins->e4sm.GUICMVList[EIV_CMDTAG_NA] = ulData[0] & 0x0000FFFF;
			ins->e4sm.opCMVS[uiCntr].ulData[0] = ulData[0];
			break;
		
		case EIV_CMDTAG_FA:
			ins->e4sm.GUICMVList[EIV_CMDTAG_FA] = ulData[0] & 0x0000FFFF;
			ins->e4sm.opCMVS[uiCntr].ulData[0] = ulData[0];
			break;
		
		case EIV_CMDTAG_LS:
			ins->e4sm.GUICMVList[EIV_CMDTAG_LS] = ulData[0] & 0x0000FFFF;
			ins->e4sm.opCMVS[uiCntr].ulData[0] = ulData[0];
			break;
		
		case EIV_CMDTAG_NFLG:
			ins->e4sm.GUICMVList[EIV_CMDTAG_NFLG] = ulData[0] & 0x0000FFFF;
			ins->e4sm.opCMVS[uiCntr].ulData[0] = ulData[0];

			if ( ulData[0] &  0x4000)
			{
				pStat->cntrPF4ResetFlag++;
								
				*pSubState = EIV_SUBSTATE_INIT;
				*pCurrState = EIV_STATE_RESET;
				return;
			}

			if ( ulData[0] &  0x0001)
			{
				pStat->cntrLOS++;
								
				*pSubState = EIV_SUBSTATE_INIT;
				*pCurrState = EIV_STATE_RESET;
				return;
			}

			break;
		
		case EIV_CMDTAG_NS:
			ins->e4sm.GUICMVList[EIV_CMDTAG_NS] = ulData[0] & 0x0000FFFF;
			ins->e4sm.opCMVS[uiCntr].ulData[0] = ulData[0];
			if ( ulData[0] &  0x8000)
			{
				pStat->cntrNegSNR++;
		
			}
			break;
		
		case EIV_CMDTAG_FS:
			ins->e4sm.GUICMVList[EIV_CMDTAG_FS] = ulData[0] & 0x0000FFFF;
			ins->e4sm.opCMVS[uiCntr].ulData[0] = ulData[0];
			break;

		
		case EIV_CMDTAG_NH:
			ins->e4sm.opCMVS[uiCntr].ulData[0] = ulData[0];
			ins->e4sm.opCMVS[uiCntr].ulData[1] = ulData[1];
			ins->e4sm.GUICMVList[EIV_CMDTAG_NH] = ulData[0] + ulData[1];

			break;
		
		case EIV_CMDTAG_NC:
			ins->e4sm.opCMVS[uiCntr].ulData[0] = ulData[0];
			ins->e4sm.GUICMVList[EIV_CMDTAG_NC] = (ulData[0] & 0xFFFF) + (( ulData[0] >> 16 ) & 0xFFFF);

			break;
		
		case EIV_CMDTAG_FC:
			ins->e4sm.opCMVS[uiCntr].ulData[0] = ulData[0];
			ins->e4sm.GUICMVList[EIV_CMDTAG_FC] = (ulData[0] & 0xFFFF) + (( ulData[0] >> 16 ) & 0xFFFF);

			break;
		
		case EIV_CMDTAG_NF:
			ins->e4sm.opCMVS[uiCntr].ulData[0] = ulData[0] & 0xFFFF;
			ins->e4sm.GUICMVList[EIV_CMDTAG_NF] = ulData[0] & 0xFFFF;

			break;
		
		case EIV_CMDTAG_FF:
			ins->e4sm.opCMVS[uiCntr].ulData[0] = ulData[0] & 0xFFFF;
			ins->e4sm.GUICMVList[EIV_CMDTAG_FF] = ulData[0] & 0xFFFF;

			break;
		
		case 	EIV_CMDTAG_DLT:
			ins->e4sm.opCMVS[uiCntr].ulData[0] = ulData[0] & 0xFFFF;
			ins->e4sm.GUICMVList[EIV_CMDTAG_DLT] = ulData[0] & 0xFFFF;

			break;
		
		case EIV_CMDTAG_DMT:
			ins->e4sm.opCMVS[uiCntr].ulData[0] = ulData[0] & 0xFFFF;
			ins->e4sm.GUICMVList[EIV_CMDTAG_DMT] = ulData[0] & 0xFFFF;

			break;

		default:

			break;
		}
		uiCntr++;
		*pSubState = EIV_SUBSTATE_TX;
		if (uiCntr<uiMaxNumOfCMV)
		{
			
		}
		else
		{
			eu_dbg (DBG_SM,"cntr<%d> reseting the cntr\n",uiCntr);
			uiCntr = ins->e4sm.usNumOnetimeOperationalCmvs;
			
		}

		ENDCASE

	default:
		break;
	}
	eu_leaves (DBG_SM);
}

