/*****************************************************************************/
/* Copyright 2006 Ikanos Communication                                       */
/*                                                                           */
/* "Ikanos USB ADSL2 Driver for Linux" is distributed in the hope that it    */
/* will be useful, but WITHOUT ANY WARRANTY; without even the implied        */
/* warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the  */
/* GNU General Public License for more details.                              */
/*                                                                           */
/* You should have received a copy of the GNU General Public License         */
/* along with "Ikanos USB ADSL2 Driver for Linux"; if not, write to the Free */
/* Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,            */
/* MA  02111-1307  USA                                                       */
/*                                                                           */
/* Written by Farinam Farahmand (farinam@gmail.com                           */
/*                                                                           */
/*****************************************************************************/

#ifndef EAGLEIV_H
#define EAGLEIV_H

#include "e4_mp.h"
/*#define IS_EAGLE_IV 1*/

#define MAX_NUM_GUI_CMVS   30


#define L1_TAG_SIZE        0x10 /* 16 byte char string.*/
#define SWAPPAGEHEADER_INDEXARRAY_SIZE    0x58
#define NO_SWAPPAGE_HEADERS     0x31 /* 49 Decimal */
#define TOTAL_SWAPPAGE_SIZE     500 * 1024 /* Max L1 size. ( have to check the size though ) */

/* 24 bytes, and is aligned to 16-byte loction */
typedef struct { 
   uint32_t ulBlockOffset;     /* file offset of the block starting from 0x500 */
   uint32_t ulNextBlockFlag;   /* If set to 1, then next header has the info of the next swap page.*/
   uint32_t ulUnCompressedCRC;
   uint32_t ulSwapPageSize;    /* assuming there is no page larger than 64KB*/
   uint32_t ulSwapPageAddress;
   uint16_t usCompressedSize;
   uint16_t usPageNumber;
} SwapPageHeader;



typedef struct {
   uint8_t strL1tag[L1_TAG_SIZE];
   uint8_t ucSwapPageHeaderIndex[SWAPPAGEHEADER_INDEXARRAY_SIZE];
   SwapPageHeader   spageHeader[NO_SWAPPAGE_HEADERS];
   uint8_t ucSwapPagesBuffer [TOTAL_SWAPPAGE_SIZE]; 
} L1_FILE, *PL1File;



typedef struct {
   uint16_t  usADIHeader;
   uint8_t   ucInitBoot;
   uint8_t   ucPageNumber;
   uint32_t  uiDspSize;
   uint32_t  uiDspAddress;
   uint16_t  usReserved;
} LOAD_EAGLEIV_PAGE_INFO;

typedef enum {
   EIV_STATE_JUST_PLUGGED_IN,
   EIV_STATE_RESET,
   EIV_STATE_SETUP,
   EIV_STATE_UNTRAINED,
   EIV_STATE_OPERATIONAL
} EIV_STATE;

typedef enum {
   EIV_SUBSTATE_INIT,
   EIV_SUBSTATE_TX,
   EIV_SUBSTATE_RX,
   EIV_SUBSTATE_DONE
} EIV_SUBSTATE;

typedef struct {
   uint16_t  cntrShowtime;
   uint16_t  cntrLOS;
   uint16_t  cntrNegSNR;
   uint16_t  cntrPF4ResetFlag;
   uint16_t  cntrSetupRXPending;
   uint16_t  cntrSetupTXPending;
   uint16_t  cntrUntrainedRXPending;
   uint16_t  cntrUntrainedTXPending;
   uint16_t  cntrUserRXPending;
   uint16_t  cntrUserTXPending;
   uint16_t  cntrOperationalTXPending;
   uint16_t  cntrOperationalRXPending;
   uint16_t  cntrCMVFailure;
   uint16_t  cntrModemReadyFailure;
   uint16_t  cntrLinkStatFaiure;
} EIV_STAT;

typedef struct {

   uint8_t   ucTotalNumPages;
   uint32_t  uiTotalDspSize;
   L1_FILE   *pDspBuff;
   LOAD_EAGLEIV_PAGE_INFO   loadPageInfo;
   
   EIV_STATE   currState;
   EIV_STATE   oldState;
   EIV_SUBSTATE   subState;
   EIV_SUBSTATE   userSubState;
   
   uint8_t   ucUserCMVRequest;
   uint8_t   ucUserResetRequest;
   uint8_t   ucRateUpdateFlag;
   uint16_t  usLinkStat;
   uint16_t  usNumSetupCmvs;
   uint16_t  usNumOperationalCmvs;
   uint16_t  usNumOnetimeOperationalCmvs;
   EIV_OPERATIONAL_CMVS  opCMVS[MAX_NUM_OPERATIONAL_CMVS];
   uint16_t  mp_retrain_sequence[MAX_NUM_RETRAIN_CMVS][EAGLEIV_MP_BUFFER_SIZE];
   uint16_t  mp_statcmv[EAGLEIV_MP_BUFFER_SIZE];
   EIV_STAT  stat;
   uint32_t  GUICMVList[MAX_NUM_GUI_CMVS];
   uint8_t   LineType;
   
   uint32_t  UpstreamRate;
   uint32_t  DownstreamRate;
   
   
/*
    uint32_t   CurrentExpirationTime;
    uint32_t   SwapPageRequiredTime;

    uint32_t  OutboundPending;
    uint32_t  InboundAsyncPending;
    uint32_t  InboundSyncPending;
*/
    uint16_t  raw_rx_msg_sync[EAGLEIV_MP_BUFFER_SIZE];
    uint16_t  raw_rx_msg_async[EAGLEIV_MP_BUFFER_SIZE];
    e4_cmv_msg_t e4_rx_cmv;
    
    uint8_t   mp_tx_msg_pending;
    uint8_t   mp_rx_async_msg_pending;
    uint8_t   mp_rx_sync_msg_pending;
    uint8_t   swap_page_request_pending;
    uint8_t   inSM;
    
#ifndef E4USER    
    struct timer_list timerSM;
#endif    
    uint32_t   HeartbeatCounter;
    uint32_t   CurrentExpirationTime;
    uint32_t   SwapPageRequiredTime;
        
} EAGLE_IV, *PEAGLE_IV;

#endif
