
/*
 *
 * Copyright (c) 2004, Frederick Ros (sl33p3r@free.fr)
 * Forked from ADI Linux driver from Analog Devices Inc.,
 *
 * Rewritten for EagleIV driver: Copyright 2006 Ikanos Communication
 										 
 * eu_boot_sm.c		 
 *										 
 * ----------------------------------------------------------------------------
 *
 * This file is part of the eagle-usb driver package.
 *
 * "eagle-usb driver package" is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * "eagle-usb driver package" is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with "eagle-usb driver package"; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * $Id: eu_boot_sm.c,v 1.4 2004/10/27 17:16:27 sleeper Exp $
 */

#include "eu_boot_sm.h"
#include "Adiutil.h"
#include "Me.h"
#include "eu_types.h"
#include "eu_utils.h"
#include "debug.h"
#include "eu_sm.h"

#define kMAILBOX_EMPTY          0
#define kMAILBOX_FULL           1

int e4_upload_idma_swapp(eu_instance_t *ins, unsigned char ucPageNumber, unsigned char ucInitialBoot);
int e4_upload_requested_p (eu_instance_t *ins, uint8_t ucPageNumber);
static int e4_upload_idma_mainp (eu_instance_t *ins);
static int upload_p ( eu_instance_t *ins );
static int eu_idma_write (eu_instance_t *ins, void *data, uint32_t size);
USB_COMPLETION_PROTO (eu_idma_complete,urb,regs);
  
void e4_uploadp( void *data )
{
    int retcode = 1;	
    eu_instance_t *ins = (eu_instance_t *)data;
    
    eu_enters (DBG_BOOT);
    
    /*
     * Check we get instance ...
     */
    if ( !ins ) 
    {
        eu_err ("Invalid instance pointer !\n");
        return;
    }
    
    retcode = upload_p (ins);  
    
    if ( retcode != 1 ) 
    {
        //should reset the Modem!
    }    
    eu_leaves (DBG_BOOT);
    
    
}

static int upload_p ( eu_instance_t *ins ) 
{
    int retcode = 1;
    
    eu_enters (DBG_BOOT);

    if ( EU_TEST_FLAG( ins, EU_DSP_IPG ) ) 
    {
        eu_report ("DSP is being uploaded. I give up.");
        retcode = 0;
        goto byebye;
    }
    if(ins->swap_data == 0xFF)
   {
     retcode = boot_the_e4modem (ins);
   }
   else
   {
    retcode = e4_upload_requested_p(ins,ins->swap_data); 
   }        
  byebye:    
    eu_leaves (DBG_BOOT);
    return (retcode);
}



/**
 * FFD boot_the_e4modem
 *
 * Does the procedures that needs to be done for Booting
 * the EIV modem
 */
 
int boot_the_e4modem (eu_instance_t *ins)
{
    int stat;

    eu_enters (DBG_BOOT);

     /*
     * To kickoff the boot process, do a custom command, SET_MODE to
     * BOOTMODE_IDMA
     
     
	stat = eu_cmd_to_modem(ins, EU_CMD_SET_MODE, MODE_BOOTMODE_IDMA, 0, NULL);
	if ( stat != 0 )
	{
	    eu_err( "eu_cmd_to_modem error %x on SET_MODE BOOTMODE_IDMA\n", stat);
	    goto end_boot;
	}
    	*/
	
	/* It's a duplicate cmd send ! try if it works without this and Wait for xxxx*/
    	
	stat = eu_cmd_to_modem(ins, EU_CMD_SET_MODE, MODE_BOOTMODE_IDMA, 0, NULL);
    	if ( stat != 0 )
    	{
        	eu_err("eu_cmd_to_modem error %x on SET_MODE BOOTMODE_IDMA\n", stat);
        	goto end_boot;
    	}
     	
	/*
     	* Advance to stage 1 and wait for 200 ms
     	*/
	
	wait_ms(200);
	
	stat = eu_cmd_to_modem(ins, EU_CMD_SET_MODE, MODE_START_RESET, 0, NULL);
    	if ( stat != 0 )
   	 {
                eu_err ("eu_cmd_to_modem error %x on SET_MODE START_RESET\n", stat);
        	goto end_boot;
    	 }

    	/*
     	* Advance to stage 2 and wait for 200 ms
     	*/
	wait_ms(200);
	stat = eu_cmd_to_modem(ins, EU_CMD_SET_MODE, MODE_END_RESET, 0, NULL);
    	if ( stat != 0 )
    	{
        	eu_err ("eu_cmd_to_modem error %x on SET_MODE END_RESET\n", stat);
        	goto end_boot;
    	}
	
	/*
     	* Advance to stage 3 and wait for 1000 ms
     	*/
	wait_ms(1000);
	
	if (e4_upload_idma_mainp(ins) == 1)
    	{
      		eu_report ("Could send EIV initial main pages -> waiting for Asynch Modem Ready\n");
    		stat = 0;
	}
    	else
    	{
        	eu_err ("Failed send EIV initial main pages \n");
        	stat = -1;
	}
	
  end_boot:
    eu_leaves (DBG_BOOT);
    return stat;
}




/**
 * FFD e4_upload_requested_p
 *
 * Calls "e4_upload_idma_swapp" to load the
 * requested EagleIV swap page
 *
 * return: 1 if everything is OK, 0 otherwise...
 */
int e4_upload_requested_p (eu_instance_t *ins, uint8_t ucPageNumber)
{

    int retcode = 1;
   
    eu_enters (DBG_BOOT);
    
    if ( EU_TEST_FLAG( ins, EU_DSP_IPG ) ) 
    {
        eu_report ("DSP is being uploaded. I give up.");
        retcode = -1;
        goto byebye;
    }
        
    /* 0 indicates requested page.*/
    if( e4_upload_idma_swapp(ins, ucPageNumber, 0) != 1)
    {
    	   eu_err ("e4_upload_requested_p: Can't send the requested swap page#%d to EagleIV.\n",ucPageNumber);
           
	   retcode = 0;
           goto byebye;
    } 
    
    
  byebye:    
    eu_leaves (DBG_BOOT);
    return (retcode);
}


/**
 * FFD e4_upload_idma_mainp
 *
 * Calls "e4_upload_idma_swapp" to load initial EagleIV
 * Swap pages 
 * return: 1 if everything is OK, 0 otherwise...
 */
static int e4_upload_idma_mainp (eu_instance_t *ins)
{

    int i;
    
    eu_enters (DBG_BOOT);
    
    eu_dbg (DBG_BOOT,"Loading Eagle IV main pages. <e4_upload_idma_swapp>\n");
           
    for (i=1; i< ins->e4sm.ucTotalNumPages; i++)
     {
     	/* 
	*  1 tells "e4_upload_idma_swapp" that these are initial swapp pages
	*/
	 if (e4_upload_idma_swapp (ins, (unsigned char)i, 1) != 1) 
         {
	   eu_err ("Can't send EagleIV initial pages.\n");
           return 0;
     	 }
	wait_ms(5);
     }
    
    /* 0xff signals 8051 that initial pages is finished.*/
    if (e4_upload_idma_swapp(ins, 0xFF, 0) != 1) 
    { 
        eu_err ("Can't send the end page of EagleIV initial pages.\n");
        return 0;
    }
    
    eu_leaves (DBG_BOOT);
    return(1);
}

/**
 * e4_upload_idma_swapp
 *
 *  Loads EagleIV Swap pages
 * 
 *  ucPageNumber: 
 *	     The swap page number to be loaded
 *
 *  ucInitialBoot: 
 *           1 if it's one of the initial swap pages
 *	     0 if it's the requested swap page
 *  return: 
 *	    1 if everything is OK, 0 otherwise...
 */
int e4_upload_idma_swapp(eu_instance_t *ins, unsigned char ucPageNumber, unsigned char ucInitialBoot)
{ 
	int stat;
	unsigned int uiBlockIndex,uiBlockOffset,uiSwapPageSize;
	unsigned int uiSize=0;
	unsigned int uiSentSize,uiUrbSize =8 * 1024;  /* suggested URB size for the USB phy & Modem IDMA boot limits*/
  	LOAD_EAGLEIV_PAGE_INFO *pLoadPageInfo;
	L1_FILE *pDspBuff = ins->e4sm.pDspBuff;
	pLoadPageInfo = &ins->e4sm.loadPageInfo;
	uiBlockIndex = pDspBuff->ucSwapPageHeaderIndex[ucPageNumber];

        eu_enters (DBG_BOOT);
	/*
	 * if page number == 0xff ==> initial load is done. 
	 */
	

	if( (ucPageNumber == 0) || ((ucPageNumber > ins->e4sm.ucTotalNumPages) && (ucPageNumber != 0xff)) )
	{  
		eu_err ("e4_upload_idma_swapp: wrong swap page number %d\n", ucPageNumber);
		return 0;
	}

	
	/*
	* 0xff indicates that the intial boot is done
	*/
	if(ucPageNumber == 0xff) 
		uiBlockIndex = 0;
	
 
        /*
	* ff indicates the locations that are not used for swap pages
	*/
	if( uiBlockIndex == 0xff)	
	{
		eu_err ("wrong block number %d\n", uiBlockIndex);
		return 0;
	}


	/*
	* there are cases that we need to load two pages consecutively. 
	* if (ulNextBlockFlag == 1), then next block holds the next page 
	* that should be loaded right after the current block.
	*/
	do
	{
		if(pDspBuff->spageHeader[uiBlockIndex].ulSwapPageSize == 0 )
		{
			eu_dbg (DBG_BOOT,"Ee4_upload_idma_swapp: ulSwapPageSize is 0, Skipping the page#%d.\n",ucPageNumber);
			return 1;
		}
		if(ucInitialBoot == 1)	
		{
			/*
			* if BIT31 is 0, then this is not an initial page.
			* bit 31 indicates initial-boot page	
			*/
			if(!(pDspBuff->spageHeader[uiBlockIndex].ulSwapPageSize & 0x80000000)) 
				return 1;
		}

		pLoadPageInfo->usADIHeader = htons(0xABCD);
		pLoadPageInfo->ucInitBoot = ucInitialBoot;
		pLoadPageInfo->ucPageNumber = ucPageNumber;
		
		/* swpapage length in bytes*/
		uiSwapPageSize = (pDspBuff->spageHeader[uiBlockIndex].ulSwapPageSize	 & 0x7fffffff) * 4; 
		
		pLoadPageInfo->uiDspSize = htonl(uiSwapPageSize);
		pLoadPageInfo->uiDspAddress = htonl(pDspBuff->spageHeader[uiBlockIndex].ulSwapPageAddress);
		pLoadPageInfo->usReserved = 0xFFFF;

				 
		 /*
         	  * Write the structure telling the DSP what to expect in the next block
         	  */
        	stat = eu_idma_write(ins, pLoadPageInfo, sizeof(LOAD_EAGLEIV_PAGE_INFO));
        
        	if ( stat != 0 ) 
        	{
            		eu_err ("Error %#X on IDMA write of LoadPageInfo\n", stat);
            		if (stat == -ENODEV) 
            		{
                		eu_err ("Wow .. this is a \"no device\" one. I give up.\n");
                		/*return -ENODEV;*/
				return 0;
            		}
        	}
		
		
		/* 0xff means the intial boot is done*/
		if(ucPageNumber == 0xff)	
		{
			eu_dbg (DBG_BOOT," EagleIV Init Page loaded ...\n");
			return 1;
		}

		/*
		* sending swap page..
		* we might need to break this to smaller pages; 8k pages are tested OK.
		*/
		
		uiBlockOffset = pDspBuff->spageHeader[uiBlockIndex].ulBlockOffset - 0x500;
		uiSentSize = 0;
		while ( uiSwapPageSize )
		{
			uiSize = ((uiSwapPageSize / uiUrbSize) == 0 ) ? uiSwapPageSize : uiUrbSize;
			
			/*
			* Now write the actual block data
         		*/
			/* FFDTEST */
			wait_ms(1);
    			stat = eu_idma_write(ins, (void *)(pDspBuff->ucSwapPagesBuffer + uiBlockOffset + uiSentSize), uiSize);
        		 if ( stat != 0 ) 
        		{
            			eu_err ("Error %#X on IDMA write of block data\n", stat);
            			if (stat == -ENODEV) 
            			{
                			eu_err ("Wow .. this is a \"no device\" one. I give up.\n");
                			
					/*return -ENODEV;*/
					return 0;
            			}
            
        		}
        	        
			/* FFDTEST*/
			wait_ms(1);
			
			eu_dbg (DBG_BOOT," uiSwapPageSize =%x\n",uiSwapPageSize);
			
			
			
			uiSwapPageSize -= uiSize;
			uiSentSize += uiSize;
		}
		eu_dbg (DBG_BOOT," page:<%d>  size<%X> address<%X> offset<%X> \n", ucPageNumber,uiSwapPageSize,
					pDspBuff->spageHeader[uiBlockIndex].ulSwapPageAddress,uiBlockOffset);
	 /* if swap page has more than one block ( pm & dm or .. ) we keep loading the next page.*/
	} while(pDspBuff->spageHeader[uiBlockIndex++].ulNextBlockFlag);
	
	eu_leaves (DBG_BOOT);
	return 1;
}

/**
 * eu_idma_write
 *
 * Responsible for sending data to the IDMA interface via the BULK IDMA OUT
 * pipe.
 *
 */
 static int eu_idma_write (eu_instance_t *ins, void *data, uint32_t size)
{

    int ret = 0;
    struct urb *urb;
    uint8_t *xfer_buf = GET_KBUFFER(size);

    eu_enters (DBG_BOOT);

    if (!xfer_buf)
    {
       eu_err ("kmalloc in eu_idma_write for %#X bytes failed\n", size);
       
       ret = -ENOMEM;
       goto IDMA_exit;
    }

    memcpy(xfer_buf, data, size);

    /*
     * Get an URB and prepare it for submission
     */
    urb = USB_ALLOC_URB ( 0 , GFP_ATOMIC );

    usb_fill_bulk_urb ( urb, ins->usbdev, ins->pipe_bulk_idma_out,
                        xfer_buf, size, eu_idma_complete, ins );
    
    urb->transfer_flags |= USB_QUEUE_BULK;

    eu_dbg (DBG_BOOT,"Submitting BULK URB %p\n", urb);
    ret = USB_SUBMIT_URB ( urb, GFP_ATOMIC );

IDMA_exit:
    eu_leaves (DBG_BOOT);
    return ret;
}

/**
 * eu_idma_complete
 *
 * Called when an eu_idma_write urb has completed. All we have to do is free
 * the memory allocated in eu_idma_write.
 *
 */

USB_COMPLETION_PROTO (eu_idma_complete,urb,regs)
{
    eu_enters (DBG_BOOT);

    /*
     * All we really need to do is free the memory we alloced and free the urb
     */
    FREE_KBUFFER(urb->transfer_buffer);
    usb_free_urb(urb);

    eu_dbg (DBG_BOOT,"Freed BULK URB %p\n", urb);
    eu_leaves (DBG_BOOT);
}
